# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
from dogtail import errors
from dogtail import predicate
from dogtail import rawinput
from dogtail import tree
from dogtail.config import config
from dogtail.logging import debug_log
from dogtail.utils import Lock

"""
Dogtail's procedural UI
All the classes here are intended to be single-instance, except for Action.
"""

__author__ = """
Zack Cerza <zcerza@redhat.com>
"""

# WARNING: Here There Be Dragons (TM)                                        #
#                                                                            #
# If you don't understand how to use this API, you almost certainly don't    #
# want to read the code first. We make use of some very non-intuitive        #
# features of Python in order to make the API very simplistic. Therefore,    #
# you should probably only read this code if you're already familiar with    #
# some of Python's advanced features. You have been warned.                  #



class FocusError(Exception):
    pass


def focusFailed(pred):
    errors.warn("The requested widget could not be focused: %s" % pred.debugName)


ENOARGS = "At least one argument is needed"


class FocusBase:
    """
    The base for every class in the module. Does nothing special, really.
    """

    node = None

    def __getattr__(self, name):
        # Fold all the Node's AT-SPI properties into the Focus object.
        try:
            return getattr(self.node, name)
        except AttributeError:
            raise AttributeError(name)


    def __setattr__(self, name, value):
        # Fold all the Node's AT-SPI properties into the Focus object.
        if name == "node":
            setattr(self.__class__, name, value)
        else:
            try:
                setattr(self.node, name, value)
            except AttributeError:
                raise AttributeError(name)


class FocusApplication(FocusBase):
    """
    Keeps track of which application is currently focused.
    """

    desktop = tree.root

    def __call__(self, name):
        """
        Search for an application that matches and refocus on the given name.
        """

        try:
            pred = predicate.IsAnApplicationNamed(name)
            app = self.desktop.findChild(pred, recursive=False, retry=False)

        except tree.SearchError:
            if config.fatalErrors:
                raise FocusError(name)
            focusFailed(pred)
            return False

        if app:
            FocusApplication.node = app
            FocusDialog.node = None
            FocusWindow.node = None
            FocusWidget.node = None

        return True


class FocusDesktop(FocusBase):
    """
    This isn't used yet, and may never be used.
    """
    pass


class FocusWindow(FocusBase):
    """
    Keeps track of which window is currently focused.
    """

    def __call__(self, name):
        """
        Search for a dialog that matches the given name and refocus on it.
        """
        result = None
        pred = predicate.IsAWindowNamed(name)

        try:
            result = FocusApplication.node.findChild(pred, requireResult=False, recursive=False)
        except AttributeError:
            pass

        if result:
            FocusWindow.node = result
            FocusDialog.node = None
            FocusWidget.node = None

        else:
            if config.fatalErrors:
                raise FocusError(pred.debugName)

            focusFailed(pred)
            return False

        return True


class FocusDialog(FocusBase):
    """
    Keeps track of which dialog is currently focused.
    """

    def __call__(self, name):
        """
        Search for a dialog that matches the given name and refocus on it.
        """
        result = None
        pred = predicate.IsADialogNamed(name)

        try:
            result = FocusApplication.node.findChild(pred, requireResult=False, recursive=False)
        except AttributeError:
            pass

        if result:
            FocusDialog.node = result
            FocusWidget.node = None

        else:
            if config.fatalErrors:
                raise FocusError(pred.debugName)

            focusFailed(pred)
            return False

        return True


class FocusWidget(FocusBase):
    """
    Keeps track of which widget is currently focused.
    """

    def findByPredicate(self, pred):
        debug_log("findByPredicate(self, pred=%s)" % str(pred))

        result = None

        try:
            result = FocusWidget.node.findChild(pred, requireResult=False, retry=False)
        except AttributeError:
            pass

        if result:
            FocusWidget.node = result
        else:
            try:
                result = FocusDialog.node.findChild(pred, requireResult=False, retry=False)
            except AttributeError:
                pass

        if result:
            FocusWidget.node = result
        else:
            try:
                result = FocusWindow.node.findChild(pred, requireResult=False, retry=False)
            except AttributeError:
                pass

        if result:
            FocusWidget.node = result

        else:
            try:
                result = FocusApplication.node.findChild(pred, requireResult=False, retry=False)
                if result:
                    FocusWidget.node = result
            except AttributeError:
                if config.fatalErrors:
                    raise FocusError(pred)

                focusFailed(pred)
                return False

        if result is None:
            FocusWidget.node = result

            if config.fatalErrors:
                raise FocusError(pred.debugName)

            focusFailed(pred)
            return False

        return True


    def __call__(self, name="", roleName="", description=""):
        """
        If name, roleName or description are specified,
        search for a widget that matches and refocus on it.
        """

        if not name and not roleName and not description:
            raise TypeError(ENOARGS)

        pred = predicate.GenericPredicate(name=name, roleName=roleName, description=description)
        return self.findByPredicate(pred)


class Focus(FocusBase):
    """
    The container class for the focused application, dialog and widget.
    """

    def __getattr__(self, name):
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in ('application', 'dialog', 'widget', 'window'):
            self.__dict__[name] = value
        else:
            raise AttributeError(name)

    desktop = tree.root
    application = FocusApplication()
    app = application  # shortcut
    dialog = FocusDialog()
    window = FocusWindow()
    frame = window
    widget = FocusWidget()

    def button(self, name):
        """
        A shortcut to self.widget.findByPredicate(predicate.IsAButtonNamed(name))
        """

        debug_log("button(self, name=%s)" % str(name))

        return self.widget.findByPredicate(predicate.IsAButtonNamed(name))


    def icon(self, name):
        """
        A shortcut to self.widget(name, roleName = 'icon')
        """

        debug_log("icon(self, name=%s)" % str(name))

        return self.widget(name=name, roleName="icon")


    def menu(self, name):
        """
        A shortcut to self.widget.findByPredicate(predicate.IsAMenuNamed(name))
        """

        debug_log("menu(self, name=%s)" % str(name))

        return self.widget.findByPredicate(predicate.IsAMenuNamed(name))


    def menuItem(self, name):
        """
        A shortcut to self.widget.findByPredicate(predicate.IsAMenuItemNamed(name))
        """

        debug_log("menuItem(self, name=%s)" % str(name))

        return self.widget.findByPredicate(predicate.IsAMenuItemNamed(name))


    def table(self, name=''):
        """
        A shortcut to self.widget(name, roleName 'table')
        """

        debug_log("table(self, name=%s)" % str(name))

        return self.widget(name=name, roleName="table")


    def tableCell(self, name=''):
        """
        A shortcut to self.widget(name, roleName 'table cell')
        """

        debug_log("tableCell(self, name=%s)" % str(name))

        return self.widget(name=name, roleName="table cell")


    def text(self, name=""):
        """
        A shortcut to self.widget.findByPredicate(IsATextEntryNamed(name))
        """

        debug_log("text(self, name=%s)" % str(name))

        return self.widget.findByPredicate(predicate.IsATextEntryNamed(name))


class Action(FocusWidget):
    """
    Aids in executing AT-SPI actions, refocusing the widget if necessary.
    """

    def __init__(self, action):
        """
        Action is a string with the same name as the AT-SPI
        action you wish to execute using this class.
        """
        self.action = action


    def __call__(self, name="", roleName="", description="", delay=config.actionDelay):
        """
        If name, roleName or description are specified, first search for
        a widget that matches and refocus on it. Then execute the action.
        """

        if name or roleName or description:
            FocusWidget.__call__(self, name=name, roleName=roleName, description=description)

        self.node.doActionNamed(self.action)


    def __getattr__(self, attr):
        return getattr(FocusWidget.node, attr)


    def __setattr__(self, attr, value):
        if attr == "action":
            self.__dict__[attr] = value
        else:
            setattr(FocusWidget, attr, value)


    def button(self, name):
        """
        A shortcut to self(name, roleName = 'push button')
        """

        debug_log("button(self, name=%s)" % str(name))

        try:
            self.__call__(name=name, roleName="push button")
        except Exception:
            self.__call__(name=name, roleName="button")


    def menu(self, name):
        """
        A shortcut to self(name, roleName = 'menu')
        """

        debug_log("menu(self, name=%s)" % str(name))

        self.__call__(name=name, roleName="menu")


    def menuItem(self, name):
        """
        A shortcut to self(name, roleName = 'menu item')
        """

        debug_log("menuItem(self, name=%s)" % str(name))

        self.__call__(name=name, roleName="menu item")


    def table(self, name=""):
        """
        A shortcut to self(name, roleName 'table')
        """

        debug_log("table(self, name=%s)" % str(name))

        self.__call__(name=name, roleName="table")


    def tableCell(self, name=""):
        """
        A shortcut to self(name, roleName 'table cell')
        """

        debug_log("tableCell(self, name=%s)" % str(name))

        self.__call__(name=name, roleName="table cell")


    def text(self, name=""):
        """
        A shortcut to self(name, roleName = 'text')
        """

        debug_log("text(self, name=%s)" % str(name))

        self.__call__(name=name, roleName="text")


class Click(Action):
    """
    A special case of Action, Click will eventually handle raw mouse events.
    """

    primary = 1
    middle = 2
    secondary = 3

    def __init__(self):
        Action.__init__(self, "click")


    def __call__(self, name="", roleName="", description="", raw=True, button=primary, delay=config.actionDelay):
        """
        By default, execute a raw mouse event. If raw is False or if button
        evaluates to False, just pass the rest of the arguments to Action.
        """

        if name or roleName or description:
            FocusWidget.__call__(self, name=name, roleName=roleName, description=description)

        if raw and button:
            Click.node.click(button)
        else:
            Action.__call__(self, name=name, roleName=roleName, description=description, delay=delay)


class Select(Action):
    """
    Aids in selecting and deselecting widgets, i.e. page tabs
    """

    select = "select"
    deselect = "deselect"

    def __init__(self, action):
        """
        action must be "select" or "deselect".
        """

        if action not in (self.select, self.deselect):
            raise ValueError(action)
        Action.__init__(self, action)


    def __call__(self, name="", roleName="", description="", delay=config.actionDelay):
        """
        If name, roleName or description are specified, first search for
        a widget that matches and refocus on it. Then execute the action.
        """

        if name or roleName or description:
            FocusWidget.__call__(self, name=name, roleName=roleName, description=description)

        func = getattr(self.node, self.action)
        func()


def type(text):
    if focus.widget.node:
        focus.widget.node.typeText(text)
    else:
        rawinput.typeText(text)


def keyCombo(combo):
    if focus.widget.node:
        focus.widget.node.keyCombo(combo)
    else:
        rawinput.keyCombo(combo)


def run(application, arguments="", appName=""):
    from dogtail.utils import run as utilsRun

    pid = utilsRun(application + " " + arguments, appName=appName)
    focus.application(application)

    return pid


if not os.path.exists("/tmp/sniff_refresh.lock"):  # pragma: no cover
    sniff_lock = Lock(lockname="sniff_refresh.lock", randomize=False, unlockOnExit=True)
    try:
        sniff_lock.lock()
    except OSError:
        pass


focus = Focus()
click = Click()
activate = Action("activate")
openItem = Action("open")
menu = Action("menu")
select = Select(Select.select)
deselect = Select(Select.deselect)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import time
import os
import pwd
import errno
import re
import subprocess
import tempfile
import random
import glob
from dogtail.config import config
from dogtail.logging import debug_log


def scratchFile(label):  # pragma: no cover
    """
    Uses tempfile.NamedTemporaryFile() to create a unique tempfile in config.scratchDir
    with a filename like: dogtail-headless-<label>.<random junk>
    """

    debug_log("scratchFile(label=%s)" % label)

    prefix = "dogtail-headless-%s." % label
    return tempfile.NamedTemporaryFile(prefix=prefix, dir=config.scratchDir, mode="w")


def testBinary(path):  # pragma: no cover
    """
    Test if given binary file exists.
    """

    debug_log("testBinary(path=%s)" % str(path))

    if (path.startswith(os.path.sep) or
            path.startswith(os.path.join(".", "")) or
            path.startswith(os.path.join("..", ""))):
        if not os.path.exists(path):
            raise IOError(errno.ENOENT, "No such file", path)

        if not os.access(path, os.X_OK):
            raise IOError(errno.ENOEXEC, "Permission denied", path)

    return True


def get_username():  # pragma: no cover
    """
    Returns a user name.
    """

    debug_log("get_username()")

    return pwd.getpwuid(os.getuid())[0]


class Subprocess:  # pragma: no cover
    """
    Subprocess class definition.
    """

    def __init__(self, cmdList, environ=None):
        testBinary(cmdList[0])
        self.cmdList = cmdList
        self.environ = environ
        self._exitCode = None


    def start(self):
        debug_log("Subprocess.start - execute the command through Popen.")

        if self.environ is None:
            self.environ = os.environ

        self.popen = subprocess.Popen(self.cmdList, env=self.environ)

        return self.popen.pid


    def wait(self):
        debug_log("Subprocess.wait - wait until Popen execution is done.")

        return self.popen.wait()


    def stop(self):
        debug_log("Subprocess.stop - terminate Popen object.")

        self.popen.terminate()


    @property
    def exitCode(self):
        debug_log("Subprocess.exitCode - returning exit code.")

        if self._exitCode is None:
            self._exitCode = self.wait()
        return self._exitCode


class XServer(Subprocess):  # pragma: no cover
    """
    XServer class definition inheriting from Subprocess class.
    """

    def __init__(self, server="/usr/bin/Xorg",
                 xinitrc="/etc/X11/xinit/Xclients",
                 resolution="1024x768x16"):
        testBinary(server)
        self.server = server
        self._exitCode = None
        self.xinit = "/usr/bin/xinit"
        self.display = None
        self.xinitrc = xinitrc
        self.resolution = resolution


    @staticmethod
    def findFreeDisplay():
        debug_log("findFreeDisplay()")

        tmp = os.listdir("/tmp")
        pattern = re.compile(".X([0-9]+)-lock")
        usedDisplays = []
        for file in tmp:
            match = re.match(pattern, file)
            if match:
                usedDisplays.append(int(match.groups()[0]))
        if not usedDisplays:
            return ":0"
        usedDisplays.sort()
        return ":" + str(usedDisplays[-1] + 1)


    @property
    def cmdList(self):
        debug_log("cmdList(self)")

        self.display = self.findFreeDisplay()

        cmd = []
        if self.xinit:
            cmd.append(self.xinit)
            if self.xinitrc:
                cmd.append(self.xinitrc)
            cmd.append("--")

        cmd.append(self.server)
        cmd.append(self.display)
        cmd.extend(["-ac", "-noreset"])

        if self.server.endswith("Xvfb"):
            cmd.extend(["-screen", "0", self.resolution])
            cmd.append("-shmem")

        return cmd


    def start(self):
        debug_log("XServer.start - starts process with command from XServer.cmdList.")

        print(' '.join(self.cmdList))
        self.popen = subprocess.Popen(self.cmdList)
        return self.popen.pid


class Script(Subprocess):  # pragma: no cover
    pass


class Session:  # pragma: no cover
    """
    Session class definition.
    """

    cookieName = "DOGTAIL_SESSION_COOKIE"

    def __init__(self, sessionBinary, scriptCmdList=[], scriptDelay=20, logout=True):
        testBinary(sessionBinary)
        self.sessionBinary = sessionBinary
        self.script = Script(scriptCmdList)
        self.scriptDelay = scriptDelay
        self.logout = logout
        self.xserver = XServer()
        self._cookie = None
        self._environment = None


    def start(self):
        debug_log("Session.start - starts a process via XServer.start")

        self.xinitrcFileObj = scratchFile("xinitrc")
        self.xserver.xinitrc = self.xinitrcFileObj.name
        self._buildXInitRC(self.xinitrcFileObj)
        xServerPid = self.xserver.start()
        time.sleep(self.scriptDelay)
        self.script.environ = self.environment
        scriptPid = self.script.start()

        return (xServerPid, scriptPid)


    @property
    def environment(self):
        debug_log("Session.environment - returns an environment")

        def isSessionProcess(fileName):
            debug_log("isSessionProcess(fileName=%s)" % str(fileName))

            try:
                if self.sessionBinary.split("/")[-1] == "startkde":
                    path_to_compare = "/usr/bin/plasma-desktop"
                else:
                    path_to_compare = self.sessionBinary

            except OSError:
                return False

            pid = fileName.split("/")[2]
            if pid == "self" or pid == str(os.getpid()):
                return False

            return True


        def getEnvDict(fileName):
            debug_log("getEnvDict(fileName=%s)" % str(fileName))

            try:
                with open(fileName, "r") as f:
                    envString = f.read()
            except IOError:
                return {}

            envItems = envString.split("\x00")
            envDict = {}

            for item in envItems:
                if "=" not in item:
                    continue
                k, v = item.split("=", 1)
                envDict[k] = v

            return envDict


        def isSessionEnv(envDict):
            debug_log("isSessionEnv(envDict=%s)" % str(envDict))
            if not envDict:
                return False

            if envDict.get(self.cookieName, "notacookie") == self.cookie:
                return True

            return False


        for path in glob.glob("/proc/*/"):
            if not isSessionProcess(path):
                continue

            envFile = path + "environ"
            envDict = getEnvDict(envFile)
            if isSessionEnv(envDict):
                self._environment = envDict

        if not self._environment:
            raise RuntimeError("Can't find our environment!")

        return self._environment


    def wait(self):
        debug_log("Session.stop - executes the XServer.wait method.")

        self.script.wait()
        return self.xserver.wait()


    def stop(self):
        debug_log("Session.stop - executes the XServer.stop method.")

        try:
            self.script.stop()
        except OSError:
            pass

        self.xserver.stop()


    def attemptLogout(self):
        debug_log("Session.attemptLogout - attempts to logout from session.")

        logoutScript = Script("dogtail-logout", environ=self.environment)
        logoutScript.start()
        logoutScript.wait()


    @property
    def cookie(self):
        debug_log("Session.cookie - returns a cookie attribute.")

        if not self._cookie:
            self._cookie = "%X" % random.getrandbits(16)
        return self._cookie


    def _buildXInitRC(self, fileObj):
        lines = [
            "export %s=%s" % (self.cookieName, self.cookie),
            "gsettings set org.gnome.desktop.interface toolkit-accessibility true",
            ". /etc/X11/xinit/xinitrc-common",
            "export %s" % self.cookieName,
            "exec -l $SHELL -c \"$CK_XINIT_SESSION $SSH_AGENT %s\"" %
            (self.sessionBinary),
            ""]

        fileObj.write("\n".join(lines).strip())
        fileObj.flush()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from dogtail.config import config
from dogtail import path
from dogtail import predicate
from dogtail import rawinput
from dogtail.rawinput import ponytail
from dogtail.logging import debug_log
from dogtail.logging import debugLogger as logger
from dogtail.utils import doDelay, Blinker, Lock
from dogtail.rawinput import SESSION_TYPE, ponytail_check_is_xwayland

from time import sleep
from types import LambdaType
import gi
from gi.repository import GLib
import os
import sys

import warnings
warnings.filterwarnings("ignore", "g_object_unref")

"""
Makes some sense of the AT-SPI API

The tree API handles various things for you:
    - fixes most timing issues
    - can automatically generate (hopefully) highly-readable logs of what the
script is doing
    - traps various UI malfunctions, raising exceptions for them (again,
hopefully improving the logs)

The most important class is Node. Each Node is an element of the desktop UI.
There is a tree of nodes, starting at 'root', with applications as its
children, with the top-level windows and dialogs as their children. The various
widgets that make up the UI appear as descendents in this tree. All of these
elements (root, the applications, the windows, and the widgets) are represented
as instances of Node in a tree (provided that the program of interest is
correctly exporting its user-interface to the accessibility system). The Node
class is a mixin for Accessible and the various Accessible interfaces.

The Action class represents an action that the accessibility layer exports as
performable on a specific node, such as clicking on it. It's a wrapper around
Accessibility.Action.

We often want to look for a node, based on some criteria, and this is provided
by the Predicate class.

Dogtail implements a high-level searching system, for finding a node (or
nodes) satisfying whatever criteria you are interested in. It does this with
a 'backoff and retry' algorithm. This fixes most timing problems e.g. when a
dialog is in the process of opening but hasn't yet done so.

If a search fails, it waits 'config.searchBackoffDuration' seconds, and then
tries again, repeatedly. After several failed attempts (determined by
config.searchWarningThreshold) it will start sending warnings about the search
to the debug log. If it still can't succeed after 'config.searchCutoffCount'
attempts, it raises an exception containing details of the search. You can see
all of this process in the debug log by setting 'config.debugSearching' to True

We also automatically add a short delay after each action
('config.defaultDelay' gives the time in seconds). We'd hoped that the search
backoff and retry code would eliminate the need for this, but unfortunately we
still run into timing issues. For example, Evolution (and probably most
other apps) set things up on new dialogs and wizard pages as they appear, and
we can run into 'setting wars' where the app resets the widgetry to defaults
after our script has already filled out the desired values, and so we lose our
values. So we give the app time to set the widgetry up before the rest of the
script runs.

The classes trap various UI malfunctions and raise exceptions that better
describe what went wrong. For example, they detects attempts to click on an
insensitive UI element and raise a specific exception for this.

Unfortunately, some applications do not set up the 'sensitive' state
correctly on their buttons (e.g. Epiphany on form buttons in a web page). The
current workaround for this is to set config.ensureSensitivity=False, which
disables the sensitivity testing.
"""

__author__ = """
Zack Cerza <zcerza@redhat.com>,
David Malcolm <dmalcolm@redhat.com>
"""

try:
    import pyatspi
    import Accessibility
except ImportError:  # pragma: no cover
    raise ImportError("Error importing the AT-SPI bindings")


if config.checkForA11y:
    from dogtail.utils import checkForA11y
    checkForA11y()


# We optionally import the bindings for libWnck.
try:
    gi.require_version("Wnck", "3.0")
    from gi.repository import Wnck
    gotWnck = True  # pragma: no cover
except (ImportError, ValueError):
    # Skip this warning, since the functionality is almost entirely nonworking anyway.
    # print "Warning: Dogtail could not import the Python bindings for
    # libwnck. Window-manager manipulation will not be available."
    gotWnck = False


haveBeenWarnedAboutActionTypes = False
haveWarnedAboutChildrenLimit = False


class SearchError(Exception):
    """
    The widget was not found.
    """


class NotSensitiveError(Exception):
    """
    The widget is not sensitive.
    """

    def __init__(self, action):
        debug_log("Not Sensitive Error Exception")

        self.message = "Cannot %s %s. It is not sensitive."
        self.action = action


    def __str__(self):
        return self.message % (self.action.name, self.action.node.getLogString())


class ActionNotSupported(Exception):
    """
    The widget does not support the requested action.
    """

    def __init__(self, actionName, node):
        debug_log("Action Not Supported Exception")

        self.message = "Cannot do '%s' action on %s"
        self.actionName = actionName
        self.node = node


    def __str__(self):
        return self.message % (self.actionName, self.node.getLogString())


class Action(object):
    """
    Class representing an action that can be performed on a specific node.
    """

    def __init__(self, node, action, index):
        self.types = ("click", "press", "release", "activate", \
            "jump", "check", "dock", "undock", "open", "menu")

        global haveBeenWarnedAboutActionTypes
        if not haveBeenWarnedAboutActionTypes:
            debug_log("Valid types of actions we know about.\n'%s'.\n" % ",".join(self.types))
            debug_log("Feel free to add any you see.")
            haveBeenWarnedAboutActionTypes = True

        self.node = node
        self.__action = action
        self.__index = index


    @property
    def name(self):
        return self.__action.getName(self.__index)


    @property
    def description(self):
        return self.__action.getDescription(self.__index)


    @property
    def keyBinding(self):
        return self.__action.getKeyBinding(self.__index)


    def __str__(self):
        return "[action | %s | %s ]" % (self.name, self.keyBinding)


    def do(self):
        """
        Performs the given tree.Action, with appropriate delays and logging.
        """

        logger.log(str("%s on %s") % (str(self.name), self.node.getLogString()))

        if not self.node.sensitive:
            if config.ensureSensitivity:
                raise NotSensitiveError(self)
            else:
                logger.log("Warning " + str(NotSensitiveError(self)))

        if config.blinkOnActions:
            self.node.blink()

        result = self.__action.doAction(self.__index)
        doDelay(config.actionDelay)

        return result


class Node(object):
    """
    A node in the tree of UI elements. This class is mixed in with
    Accessibility.Accessible to both make it easier to use and to add
    additional functionality. It also has a debugName which is set up
    automatically when doing searches.
    """

    def __setupUserData(self):
        """
        Setup user data dictionary.
        """
        try:
            len(self.user_data)
        except (AttributeError, TypeError):
            self.user_data = {}


    @property
    def debugName(self):
        """
        Debug name assigned during search operations.
        """

        self.__setupUserData()
        return self.user_data.get("debugName", None)


    @debugName.setter
    def debugName(self, debugName):
        """
        Debug name setter.
        """

        self.__setupUserData()
        self.user_data["debugName"] = debugName


    @property
    def dead(self):
        """
        Is the node dead (defunct)?
        """

        debug_log("dead(self)")

        try:
            if self.roleName == "invalid":
                return True
            assert self.role is not None
            assert self.name is not None
            if len(self) > 0:
                assert self.children[0] is not None
        except Exception:
            return True
        return False


    @property
    def children(self):
        """
        A list of this Accessible's children.
        """

        debug_log("children(self)")

        if self.parent and self.parent.roleName == "hyper link":
            debug_log(self.parent.role)
            return []

        children_list = []
        childCount = self.childCount
        if childCount > config.childrenLimit:
            global haveWarnedAboutChildrenLimit
            if not haveWarnedAboutChildrenLimit:
                logger.log("Only returning %s children. You may change "
                           "config.childrenLimit if you wish. This message will only"
                           " be printed once." % str(config.childrenLimit))
                haveWarnedAboutChildrenLimit = True
                childCount = config.childrenLimit

        for i in range(childCount):
            """
            Workaround for GNOME bug #465103
            also solution for GNOME bug #321273
            """

            try:
                child = self[i]
            except LookupError:
                child = None

            if child:
                children_list.append(child)

        invalidChildren = childCount - len(children_list)
        if invalidChildren and config.debugSearching:
            logger.log(str("Skipped %s invalid children of %s") %
                       (invalidChildren, str(self)))

        try:
            ht = self.queryHypertext()
            for li in range(ht.getNLinks()):
                link = ht.getLink(li)
                for ai in range(link.nAnchors):
                    child = link.getObject(ai)
                    if child == self:
                        continue
                    child.__setupUserData()
                    child.user_data['linkAnchor'] = \
                        LinkAnchor(node=child,
                                   hypertext=ht,
                                   linkIndex=li,
                                   anchorIndex=ai)
                    children_list.append(child)
        except (NotImplementedError, AttributeError):
            pass

        return children_list


    roleName = property(Accessibility.Accessible.getRoleName)
    role = property(Accessibility.Accessible.getRole)
    name = property(Accessibility.Accessible.name)
    parent = property(Accessibility.Accessible.parent)
    indexInParent = property(Accessibility.Accessible.getIndexInParent)
    __window_id = None


    @property
    def window_id(self):
        """
        Return window id of a node.
        """

        debug_log("window_id(self)")

        if SESSION_TYPE == "x11":
            return None

        if self.__window_id is None:
            debug_log("Window id event.")

            window_list = ponytail.window_list
            if len(window_list) == 0:
                doDelay(config.actionDelay)
                window_list = ponytail.window_list

            for window in window_list:
                if "title" not in window.keys():
                    window["title"] = ""

            node = self
            parent_list = [node]
            while node.parent is not None:
                parent_list.append(node.parent)
                node = node.parent

            for ancestor in parent_list:
                if ancestor.parent is None:
                    self.__window_id = [x["id"] for x in window_list if bool(x["has-focus"]) is True][0]
                    return [x["id"] for x in window_list if bool(x["has-focus"]) is True][0]

                elif ancestor.parent.roleName == "application" and ancestor.parent.name == "gnome-shell":
                    return ""

                elif ancestor.parent.roleName == "application" and ancestor.roleName == "window" and ancestor.name == "":
                    self.__window_id = [x["id"] for x in window_list if bool(x["has-focus"]) is True][0]
                    return [x["id"] for x in window_list if bool(x["has-focus"]) is True][0] # context menus

                elif ancestor.parent.roleName == "application" and ancestor.name in [x["title"] for x in window_list]:
                    self.__window_id = [x["id"] for x in window_list if x["title"] == ancestor.name][0]
                    return [x["id"] for x in window_list if x["title"] == ancestor.name][0]

                elif ancestor.parent.roleName == "application":
                    self.__window_id = [x["id"] for x in window_list if bool(x["has-focus"]) is True][0]
                    return [x["id"] for x in window_list if bool(x["has-focus"]) is True][0]

        else:
            return self.__window_id


    @property
    def window_has_focus(self):
        """
        Check if window is focused.
        """

        debug_log("window_has_focus(self)")

        if SESSION_TYPE == "wayland":
            window_id = self.window_id

            window_list = ponytail.window_list
            for window in window_list:
                if window["id"] == window_id and bool(window["has-focus"]) is True:
                    return True

            return False

        else:
            raise NotImplementedError


    def doActionNamed(self, name):
        """
        Perform the action with the specified name. For a list of actions
        supported by this instance, check the 'actions' property.
        """

        debug_log("doActionNamed(self, name=%s)" % str(name))

        actions = self.actions
        if name in actions:
            return actions[name].do()

        raise ActionNotSupported(name, self)


    @property
    def actions(self):
        """
        A dictionary of supported action names as keys, with Action objects as
        values. Common action names include:
            'click' 'press' 'release' 'activate' 'jump'
            'check' 'dock' 'undock' 'open' 'menu'
        """

        debug_log("actions(self)")

        actions = {}

        try:
            action_query = self.queryAction()
            for index in range(action_query.nActions):
                action_object = Action(self, action_query, index)
                actions[action_query.getName(index)] = action_object
        finally:
            debug_log("If actions are empty there could have been an exception.")
            return actions


    @property
    def combovalue(self):
        """
        The value (as a string) currently selected in the combo box.
        """

        debug_log("combovalue(self)")

        return self.name


    @combovalue.setter
    def combovalue(self, value):
        logger.log(str("Setting combobox %s to '%s'") % (self.getLogString(), str(value)))

        self.childNamed(childName=value).doActionNamed("click")
        doDelay()


    @property
    def URI(self):
        """
        The value of user data linkAnchor.
        """

        debug_log("URI(self)")

        try:
            return self.user_data["linkAnchor"].URI
        except (KeyError, AttributeError):
            raise NotImplementedError


    @property
    def text(self):
        """
        For instances with an AccessibleText interface, the text as a
        string. This is read-only, unless the instance also has an
        AccessibleEditableText interface. In this case, you can write values
        to the attribute. This will get logged in the debug log, and a delay
        will be added.

        If this instance corresponds to a password entry, use the passwordText
        property instead.
        """

        debug_log("text(self)")

        try:
            return self.queryText().getText(0, self.get_character_count())
        except NotImplementedError:
            return None


    @text.setter
    def text(self, text):
        try:
            if config.debugSearching:
                setter_message = "Setting text of %s to '%s'"

                debug_log("Maximum length is set at 140, cutting the text after.")
                if len(text) > 140:
                    txt = text[:134] + " [...]"
                else:
                    txt = text

                logger.log(str(setter_message) % (self.getLogString(), str("'%s'") % str(txt)))

            self.queryEditableText().setTextContents(text)
        except NotImplementedError:
            raise AttributeError("can't set attribute")


    @property
    def caretOffset(self):
        """
        For instances with an AccessibleText interface, the caret offset as an integer.
        """

        debug_log("caretOffset(self)")

        return self.queryText().caretOffset


    @caretOffset.setter
    def caretOffset(self, offset):
        """
        Caret offset setter.
        """

        debug_log("caretOffset(self, offset=%s)" % str(offset))

        return self.queryText().setCaretOffset(offset)


    @property
    def position(self):
        """
        A tuple containing the position of the Accessible: (x, y)
        """

        debug_log("position(self)")

        # Always start with DESKTOP_COORDS to detect GTK4 app
        pos = self.queryComponent().getPosition(pyatspi.DESKTOP_COORDS)

        # Check if it's a GTK4 app by detecting the initial position as (0,0)
        if pos == (0,0):
            if self.name == "Activities" and (self.roleName == "label" or self.roleName == "toggle button"):
                # specil case of the Activities in the corner... will not be gtk4
                return pos
            gtk4Offset = config.gtk4Offset
            # Determine if *this* node is a part of a full-screen frame
            fullscreen_offset = (0, 0) # No offset fullscreen, shadows or not!
            node = self  # Assuming 'self' is the current node
            try:
                from dogtail.utils import get_screen_resolution
                screen_width, _ = get_screen_resolution()
                while node:
                    if node.roleName == "frame":
                        frame_width, frame_height = node.size
                        if frame_width == screen_width:
                            gtk4Offset = fullscreen_offset
                        break
                    node = node.parent  # Traverse up to check for frame ancestors
            except Exception:
                pass
            # For x11 session, calculate position using WINDOW_COORDS, add up with window possition
            # and finally apply offset if not fullscreen
            if SESSION_TYPE == "x11":
                pos = self.queryComponent().getPosition(pyatspi.WINDOW_COORDS)
                from dogtail.utils import get_current_x_window_position
                base_x, base_y = get_current_x_window_position()
                # Add both gtk4Offset and the current x window position
                pos = (pos[0] + base_x + gtk4Offset[0], pos[1] + base_y + gtk4Offset[1])
            else:
                # For wayland, if it's still (0,0) return that, otherwise adjust offset directly
                pos = self.queryComponent().getPosition(pyatspi.WINDOW_COORDS)
                if pos != (0,0):
                    pos = (pos[0] + gtk4Offset[0], pos[1] + gtk4Offset[1])

        # Return the position as determined above
        return pos


    @property
    def size(self):
        """
        A tuple containing the size of the Accessible: (w, h)
        """

        debug_log("size(self)")

        if SESSION_TYPE == "wayland":
            if self.roleName == "window" or self.roleName == "dialog" or self.roleName == "frame":
                window_list = ponytail.window_list
                window_id = self.window_id
                for window in window_list:
                    if window["id"] == window_id:
                        return (int(window["width"]), int(window["height"]))

        return self.queryComponent().getSize()


    @property
    def extents(self):
        """
        A tuple containing the location and size of the Accessible: (x, y, w, h)
        """

        debug_log("extents(self)")

        try:
            ex = self.queryComponent().getExtents(pyatspi.DESKTOP_COORDS)
            if (ex.x, ex.y) == (0,0):
                ex = self.queryComponent().getExtents(pyatspi.WINDOW_COORDS)
                pos = self.position
                return (pos[0], pos[1], ex.width, ex.height)
            return (ex.x, ex.y, ex.width, ex.height)
        except NotImplementedError:
            return None

    @property
    def center(self):
        """
        A tuple containing the center position of the Accessible:(x, y)
        """

        debug_log("center(self)")

        x, y, w, h = self.extents
        return x + w / 2, y + h / 2

    def contains(self, x, y):
        """
        Can tell if and x and y coordinates are inside a desktop coordinates.
        """

        debug_log("contains(self, x=%s, y=%s)" % (str(x), str(y)))

        try:
            return self.queryComponent().contains(x, y, pyatspi.DESKTOP_COORDS)
        except NotImplementedError:
            return False


    def getChildAtPoint(self, x, y):
        """
        Get a child in coordinates x, y.
        """

        debug_log("getChildAtPoint(self, x=%s, y=%s)" % (str(x), str(y)))

        node = self
        while True:
            try:
                child = node.queryComponent().getAccessibleAtPoint(x, y, pyatspi.DESKTOP_COORDS)
                if child and child.contains(x, y):
                    node = child
                else:
                    break
            except NotImplementedError:
                break

        if node and node.contains(x, y):
            return node
        else:
            return None


    def grabFocus(self):
        """
        Attempts to set the keyboard focus to this Accessible.
        Affected by rhbz 1656447 on Wayland! We do different actions based on type of the node
        to get it focused there as workaround. For some we can do nothing (push button)
        """

        debug_log("grabFocus(self)")

        if SESSION_TYPE == "x11" or ponytail_check_is_xwayland(self.window_id) or self.window_has_focus:
            return self.queryComponent().grabFocus()

        else:
            if "toggle" in self.roleName or "check box" in self.roleName:
                self.doubleClick()

            elif "text" in self.roleName or "table cell" in self.roleName:
                self.click()

            elif "menu item" in self.roleName and self.position[0] > 0:
                self.select()

            elif "menu item" in self.roleName and self.findAncestor(predicate.GenericPredicate(roleName="combo box")):
                self.doActionNamed("click")

            else:
                pass


    def click(self, button=1):
        """
        Generates a raw mouse click event, using the specified button.
            - 1 is left,
            - 2 is middle,
            - 3 is right.
        """

        logger.log(str("Clicking on %s") % self.getLogString())

        clickX = self.position[0] + self.size[0] / 2
        clickY = self.position[1] + self.size[1] / 2


        if ("menu item" in self.roleName or self.roleName == "menu") and \
                self.parent.roleName != "menu bar" and \
                "click" in self.actions:
            # self.select()

            # doDelay(config.typingDelay)
            # window_id = None
            if self.name.lower() in ["quit", "exit"] or "close" in self.name.lower():
                window_id = ""

            ### this old workaround with select and enter was needed due to click action
            ### crashing app on these items... that is no longer an issue
            self.doActionNamed('click')
            # rawinput.pressKey("enter", window_id)
            doDelay(config.typingDelay)

        else:
            if config.debugSearching:
                logger.log(str("raw click on %s %s at (%s,%s)") %
                    (str(self.name), self.getLogString(), str(clickX), str(clickY)))
            debug_log("click(self, button=%s)" % str(button))

            rawinput.click(clickX, clickY, button, window_id=self.window_id)


    def doubleClick(self, button=1):
        """
        Generates a raw mouse double-click event, using the specified button.
        """

        clickX = self.position[0] + self.size[0] / 2
        clickY = self.position[1] + self.size[1] / 2

        if config.debugSearching:
            logger.log(str("raw click on %s %s at (%s,%s)") %
                       (str(self.name), self.getLogString(), str(clickX), str(clickY)))
        debug_log("doubleClick(self, button=%s)" % str(button))

        rawinput.doubleClick(clickX, clickY, button, window_id=self.window_id)


    def point(self, mouseDelay=None):
        """
        Move mouse cursor to the center of the widget.
        """

        pointX = self.position[0] + self.size[0] / 2
        pointY = self.position[1] + self.size[1] / 2

        logger.log(str("Pointing on %s %s at (%s,%s)") %
                   (str(self.name), self.getLogString(), str(pointX), str(pointY)))
        debug_log("point(self, mouseDelay=%s)" % str(mouseDelay))

        rawinput.point(pointX, pointY, window_id=self.window_id)


    @property
    def labeler(self):
        """
        'labeller' (read-only list of Node instances):
        The node(s) that is/are a label for this node. Generated from 'relations'.
        """

        debug_log("labeler(self)")

        relationSet = self.getRelationSet()
        for relation in relationSet:
            if relation.getRelationType() == pyatspi.RELATION_LABELLED_BY:
                if relation.getNTargets() == 1:
                    return relation.getTarget(0)

                targets = []
                for i in range(relation.getNTargets()):
                    targets.append(relation.getTarget(i))

                return targets

    labeller = labeler


    @property
    def labelee(self):
        """
        'labellee' (read-only list of Node instances):
        The node(s) that this node is a label for. Generated from 'relations'.
        """

        debug_log("labelee(self)")

        relationSet = self.getRelationSet()
        for relation in relationSet:
            if relation.getRelationType() == pyatspi.RELATION_LABEL_FOR:
                if relation.getNTargets() == 1:
                    return relation.getTarget(0)

                targets = []
                for i in range(relation.getNTargets()):
                    targets.append(relation.getTarget(i))

                return targets


    labellee = labelee


    @property
    def sensitive(self):
        """
        Is the Accessible sensitive (i.e. not greyed out)?
        """

        debug_log("sensitive(self)")

        return self.getState().contains(pyatspi.STATE_SENSITIVE)


    @property
    def showing(self):
        """
        Is the Accessible showing (rendered and visible) on the screen?
        """

        debug_log("showing(self)")

        return self.getState().contains(pyatspi.STATE_SHOWING)


    @property
    def focusable(self):
        """
        Is the Accessible capable of having keyboard focus?
        """

        debug_log("focusable(self)")

        return self.getState().contains(pyatspi.STATE_FOCUSABLE)


    @property
    def focused(self):
        """
        Does the Accessible have keyboard focus?
        """

        debug_log("focused(self)")

        return self.getState().contains(pyatspi.STATE_FOCUSED)


    @property
    def checked(self):
        """
        Is the Accessible a checked checkbox?
        """

        debug_log("Node.checked - is the Accessible a checked checkbox?")

        return self.getState().contains(pyatspi.STATE_CHECKED)


    @property
    def isChecked(self):
        """
        Is the Accessible a checked checkbox? Compatibility property, same as Node.checked.
        """

        debug_log("isChecked(self)")

        return self.checked


    @property
    def visible(self):
        """
        Is the Accessible set to be visible? A widget with set attribute
        'visible' is supposed to be shown and doesn't need to be actually
        rendered. On the other hand, a widget with unset attribute 'visible'
        """

        debug_log("visible(self)")

        return self.getState().contains(pyatspi.STATE_VISIBLE)


    def selectAll(self):
        """
        Selects all children.
        """

        debug_log("selectAll(self)")

        result = self.querySelection().selectAll()
        doDelay()
        return result


    def deselectAll(self):
        """
        Deselects all selected children.
        """

        debug_log("deselectAll(self)")

        result = self.querySelection().clearSelection()
        doDelay()
        return result


    def select(self):
        """
        Selects the Accessible.
        """

        debug_log("select(self)")

        try:
            parent = self.parent
        except AttributeError:
            raise NotImplementedError

        result = parent.querySelection().selectChild(self.indexInParent)
        doDelay()
        return result


    def deselect(self):
        """
        Deselects the Accessible.
        """

        debug_log("deselect(self)")

        try:
            parent = self.parent
        except AttributeError:
            raise NotImplementedError

        result = parent.querySelection().deselectChild(self.indexInParent)
        doDelay()
        return result


    @property
    def isSelected(self):
        """
        Is the Accessible selected? Compatibility property, same as Node.selected.
        """

        debug_log("isSelected(self)")

        try:
            parent = self.parent
        except AttributeError:
            raise NotImplementedError
        return parent.querySelection().isChildSelected(self.indexInParent)


    @property
    def selected(self):
        """
        Is the Accessible selected?
        """

        debug_log("selected(self)")

        return self.isSelected


    @property
    def selectedChildren(self):
        """
        Returns a list of children that are selected.
        """

        debug_log("selectedChildren(self)")

        selection = self.querySelection()

        selectedChildren = []
        for i in range(selection.nSelectedChildren):
            selectedChildren.append(selection.getSelectedChild(i))


    @property
    def value(self):
        """
        The value contained by the AccessibleValue interface.
        """

        debug_log("value(self)")

        try:
            return self.queryValue().currentValue
        except NotImplementedError:
            pass


    @value.setter
    def value(self, value):
        """
        Setter for the value contained by the AccessibleValue interface.
        """

        debug_log("value(self, value=%s)" % str(value))

        self.queryValue().currentValue = value


    @property
    def minValue(self):
        """
        The minimum value of Node.value
        """

        debug_log("minValue(self)")

        try:
            return self.queryValue().minimumValue
        except NotImplementedError:
            pass


    @property
    def minValueIncrement(self):
        """
        The minimum value increment of the Node.value
        """

        debug_log("minValueIncrement(self)")

        try:
            return self.queryValue().minimumIncrement
        except NotImplementedError:
            pass


    @property
    def maxValue(self):
        """
        The maximum value of self.value
        """

        debug_log("maxValue(self)")

        try:
            return self.queryValue().maximumValue
        except NotImplementedError:
            pass


    def typeText(self, string):
        """
        Type the given text into the node, with appropriate delays and logging.
        """

        logger.log(str("Typing text into %s: '%s'") % (self.getLogString(), str(string)))
        debug_log("typeText(self, string=%s)" % string)

        if self.focusable:
            if not self.focused:
                try:
                    self.grabFocus()
                except Exception:
                    logger.log("Node is focusable but I can't grabFocus!")

            rawinput.typeText(string)

        else:
            logger.log("Node is not focusable; falling back to inserting text")
            et = self.queryEditableText()
            et.insertText(self.caretOffset, string, len(string))
            self.caretOffset += len(string)
            doDelay()


    def keyCombo(self, comboString):
        """
        Press keys in combination.
        """

        if config.debugSearching:
            logger.log(str("Pressing keys '%s' into %s") %
                       (str(comboString), self.getLogString()))

        debug_log("keyCombo(self, comboString=%s)" % str(comboString))

        if self.focusable:
            if not self.focused:
                try:
                    self.grabFocus()
                except Exception:
                    logger.log("Node is focusable but I can't grabFocus!")
        else:
            logger.log("Node is not focusable; trying key combo anyway")

        rawinput.keyCombo(comboString)


    def getLogString(self):
        """
        Get a string describing this node for the logs,
        respecting the config.absoluteNodePaths boolean.
        """

        debug_log("getLogString(self)")

        if config.absoluteNodePaths:
            return self.getAbsoluteSearchPath()
        else:
            return str(self)


    def satisfies(self, pred):
        """
        Does this node satisfy the given predicate?
        """

        debug_log("satisfies(self, pred=%s)" % (str(pred)))

        assert isinstance(pred, predicate.Predicate)
        return pred.satisfiedByNode(self)


    def dump(self, type="plain", fileName=None):
        """
        Dumping a tree structure of the node.
        """

        debug_log("dump(self, type=%s, fileName=%s)" % (str(type), str(fileName)))

        from dogtail import dump
        dumper = getattr(dump, type)
        dumper(self, fileName)


    def getAbsoluteSearchPath(self):
        """
        Generate a SearchPath instance giving the 'best'
        way to find the Accessible wrapped by this node again, starting
        at the root and applying each search in turn.

        This is somewhat analagous to an absolute path in a filesystem,
        except that some of searches may be recursive, rather than just
        searching direct children.

        Used by the recording framework for identifying nodes in a
        persistent way, independent of the style of script being
        written.
        """

        debug_log("getAbsoluteSearchPath(self)")

        if config.debugSearchPaths:
            logger.log("getAbsoluteSearchPath(%s)" % self)

        if self.roleName == "application":
            result = path.SearchPath()
            result.append(predicate.IsAnApplicationNamed(self.name), False)
            return result

        else:
            if self.parent:
                (ancestor, pred, isRecursive) = self.getRelativeSearch()
                if config.debugSearchPaths:
                    debug_log("Found ancestor: %s" % ancestor)

                ancestorPath = ancestor.getAbsoluteSearchPath()
                ancestorPath.append(pred, isRecursive)
                return ancestorPath

            else:
                # This should be the root node:
                return path.SearchPath()


    def getRelativeSearch(self):
        """
        Get a (ancestorNode, predicate, isRecursive) triple that
        identifies the best way to find this Node uniquely.
        """

        debug_log("getRelativeSearch(self)")

        if config.debugSearchPaths:
            logger.log("getRelativeSearchPath(%s)" % self)

        assert self
        assert self.parent

        isRecursive = False
        ancestor = self.parent

        while not self.__nodeIsIdentifiable(ancestor):
            debug_log("Node is not identifiable, setting isRecursive as True.")
            ancestor = ancestor.parent
            isRecursive = True

        # I think this is useless, since only change is the debug string

        # Pick the most appropriate predicate for finding this node:
        if self.labellee:
            if self.labellee.name:
                return (ancestor, predicate.IsLabelledAs(self.labellee.name), isRecursive)

        if self.roleName == "menu":
            return (ancestor, predicate.IsAMenuNamed(self.name), isRecursive)
        elif self.roleName == "menu item" or self.roleName == "check menu item":
            return (ancestor, predicate.IsAMenuItemNamed(self.name), isRecursive)
        elif self.roleName == "text":
            return (ancestor, predicate.IsATextEntryNamed(self.name), isRecursive)
        elif self.roleName == "button":
            return (ancestor, predicate.IsAButtonNamed(self.name), isRecursive)
        elif self.roleName == "push button":
            return (ancestor, predicate.IsAButtonNamed(self.name), isRecursive)
        elif self.roleName == "frame":
            return (ancestor, predicate.IsAWindowNamed(self.name), isRecursive)
        elif self.roleName == "dialog":
            return (ancestor, predicate.IsADialogNamed(self.name), isRecursive)
        else:
            pred = predicate.GenericPredicate(name=self.name, roleName=self.roleName)
            return (ancestor, pred, isRecursive)


    def __nodeIsIdentifiable(self, ancestor):
        """
        Checks if given node can be identified by labellee, name or parent.
        """

        debug_log("__nodeIsIdentifiable(self, ancestor=%s)" % str(ancestor))

        if ancestor.labellee:
            return True

        elif ancestor.name:
            return True

        elif not ancestor.parent:
            return True

        else:
            return False


    def _fastFindChild(self, pred, recursive=True, showingOnly=None):
        """
        Searches for an Accessible using methods from pyatspi.utils
        """

        debug_log("_fastFindChild(self, pred=%s, recursive=%s, recursive=%s)" %
                      (str(pred), str(recursive), str(recursive)))

        if isinstance(pred, predicate.Predicate):
            pred = pred.satisfiedByNode

        if showingOnly is None:
            showingOnly = config.searchShowingOnly

        if showingOnly:
            original_predicate = pred
            pred = lambda x: original_predicate(x) and x.getState().contains(pyatspi.STATE_SHOWING)

        if not recursive:
            cIter = iter(self)
            while True:
                try:
                    child = next(cIter)
                except StopIteration:
                    break

                if child is not None and pred(child):
                    return child

        else:
            return pyatspi.utils.findDescendant(self, pred)


    def findChild(self, pred, recursive=True, debugName=None, retry=True, requireResult=True, showingOnly=None):
        """
        Search for a node satisyfing the predicate, returning a Node.

        If retry is True (the default), it makes multiple attempts, backing off and retrying
        on failure, and eventually raises a descriptive exception if the search fails.

        If retry is False, it gives up after one attempt.

        If requireResult is True (the default), an exception is raised after all
        attempts have failed. If it is false, the function simply returns None.
        """

        debug_log("findChild(self, pred=%s, recursive=%s, debugName=%s, retry=%s, requireResult=%s, showingOnly=%s)" %
                      (str(pred), str(recursive), str(debugName), str(retry), str(requireResult), str(showingOnly)))

        def describeSearch(parent, pred, recursive, debugName):
            """
            Internal helper function
            """

            noun = "descendant" if recursive else "child"

            if debugName is None:
                debugName = pred.describeSearchResult()

            return "%s of %s: %s" % (str(noun), parent.getLogString(), str(debugName))


        compare_function = None
        if isinstance(pred, LambdaType):
            compare_function = pred
            if debugName is None:
                debugName = "child satisyfing a custom lambda function"

        else:
            assert isinstance(pred, predicate.Predicate)
            compare_function = pred.satisfiedByNode

        number_of_attempts = 0
        while number_of_attempts < config.searchCutoffCount:
            if number_of_attempts >= config.searchWarningThreshold or config.debugSearching:
                logger.log(str("Searching for %s (attempt %i)") %
                           (describeSearch(self, pred, recursive, debugName), number_of_attempts))

            result = self._fastFindChild(compare_function, recursive, showingOnly=showingOnly)

            if result:
                assert isinstance(result, Node)
                if debugName:
                    result.debugName = debugName
                else:
                    result.debugName = pred.describeSearchResult()
                return result

            else:
                if not retry:
                    break
                number_of_attempts += 1
                if config.debugSearching or config.debugSleep:
                    logger.log("sleeping for '%f'" % config.searchBackoffDuration)
                sleep(config.searchBackoffDuration)

        if requireResult:
            raise SearchError(describeSearch(self, pred, recursive, debugName))


    def findChildren(self, pred, recursive=True, isLambda=False, showingOnly=None):
        """
        Find all children/descendents satisfying the predicate. You can also use lambdas in
        place of the pred that will enable search also against pure dogtail Node properties
        (like showing). I.e: "lambda x: x.roleName == 'menu item' and x.showing is True".
        isLambda does not have to be set, it's kept only for api compatibility.
        """

        debug_log("findChildren(self, pred=%s, recursive=%s, isLambda=%s, showingOnly=%s)" %
                      (str(pred), str(recursive), str(isLambda), str(showingOnly)))

        compare_function = None
        if isLambda is True or isinstance(pred, LambdaType):
            compare_function = pred

        else:
            assert isinstance(pred, predicate.Predicate)
            compare_function = pred.satisfiedByNode

        if showingOnly is None:
            showingOnly = config.searchShowingOnly

        if showingOnly:
            original_compare_func = compare_function
            compare_function = lambda n: original_compare_func(n) and \
                n.getState().contains(pyatspi.STATE_SHOWING)

        results = []
        number_of_attempts = 0
        while number_of_attempts < config.searchCutoffCount:
            if number_of_attempts >= config.searchWarningThreshold or config.debugSearching:
                logger.log("Warning: a11y errors caught, making attempt %i" % number_of_attempts)

            try:
                if recursive:
                    results = pyatspi.utils.findAllDescendants(self, compare_function)
                else:
                    results = list(filter(compare_function, self.children))
                break

            except (GLib.GError, TypeError):
                number_of_attempts += 1
                if number_of_attempts == config.searchCutoffCount:
                    logger.log("Warning: errors caught from the a11y tree, giving up search")
                    raise RuntimeError("Error: Session has probaly broken a11y! Exiting allowing session re-runs")
                else:
                    sleep(config.searchBackoffDuration)
                continue

        return results


    def findAncestor(self, pred, showingOnly=None):
        """
        Search up the ancestry of this node, returning the first Node
        satisfying the predicate, or None.
        """

        debug_log("findAncestor(self, pred=%s, showingOnly=%s)" %
                      (str(pred), str(showingOnly)))

        assert isinstance(pred, predicate.Predicate)

        candidate = self.parent
        while candidate is not None:
            if candidate.satisfies(pred):
                return candidate
            else:
                candidate = candidate.parent

        return None

    # Various wrapper/helper search methods:
    def child(self, name="", roleName="", description="", label="", identifier="", recursive=True, retry=True, debugName=None, showingOnly=None):
        """
        Finds a child satisying the given criteria.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("child(self, name=%s, roleName=%s, description=%s, label=%s, identifier=%s, recursive=%s, retry=%s, debugName=%s, showingOnly=%s)" %
            (str(name), str(roleName), str(description), str(label), str(identifier), str(recursive), str(retry), str(debugName), str(showingOnly)))

        return self.findChild(predicate.GenericPredicate(name=name, roleName=roleName, description=description,
                              label=label, identifier=identifier), recursive=recursive, retry=retry, debugName=debugName, showingOnly=showingOnly)

    def isChild(self, name="", roleName="", description="", label="", identifier="", recursive=True, retry=False, debugName=None, showingOnly=None):
        """
        Determines whether a child satisying the given criteria exists.

        This is implemented using findChild, but will not automatically retry
        if no such child is found. To make the function retry multiple times set retry to True.
        Returns a boolean value depending on whether the child was eventually found. Similar to
        'child', yet it catches SearchError exception to provide for False results, will raise
        any other exceptions. It also logs the search.
        """

        debug_log("isChild(self, name=%s, roleName=%s, description=%s, label=%s, identifier=%s, recursive=%s, retry=%s, debugName=%s, showingOnly=%s)" %
            (str(name), str(roleName), str(description), str(label), str(identifier), str(recursive), str(retry), str(debugName), str(showingOnly)))

        try:
            self.findChild(
                predicate.GenericPredicate(
                    name=name, roleName=roleName, description=description, label=label, identifier=identifier),
                recursive=recursive, retry=retry, debugName=debugName, showingOnly=showingOnly)
            return True
        except SearchError:
            return False


    def menu(self, menuName, recursive=True, showingOnly=None):
        """
        Search below this node for a menu with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("menu(self, nameName=%s, recursive=%s, showingOnly=%s)" %
                      (str(menuName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsAMenuNamed(menuName=menuName), recursive, showingOnly=showingOnly)


    def menuItem(self, menuItemName, recursive=True, showingOnly=None):
        """
        Search below this node for a menu item with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("menuItem(self, menuItemName=%s, recursive=%s, showingOnly=%s)" %
                      (str(menuItemName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsAMenuItemNamed(menuItemName=menuItemName), recursive, showingOnly=showingOnly)


    def textentry(self, textEntryName, recursive=True, showingOnly=None):
        """
        Search below this node for a text entry with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("textentry(self, textEntryName=%s, recursive=%s, showingOnly=%s)" %
                      (str(textEntryName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsATextEntryNamed(textEntryName=textEntryName), recursive, showingOnly=showingOnly)


    def button(self, buttonName, recursive=True, showingOnly=None):
        """
        Search below this node for a button with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("button(self, buttonName=%s, recursive=%s, showingOnly=%s)" %
                      (str(buttonName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsAButtonNamed(buttonName=buttonName), recursive, showingOnly=showingOnly)


    def childLabelled(self, labelText, recursive=True, showingOnly=None):
        """
        Search below this node for a child labelled with the given text.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("childLabelled(self, labelText=%s, recursive=%s, showingOnly=%s)" %
                      (str(labelText), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsLabelledAs(labelText), recursive, showingOnly=showingOnly)


    def childNamed(self, childName, recursive=True, showingOnly=None):
        """
        Search below this node for a child with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("childNamed(self, childName=%s, recursive=%s, showingOnly=%s)" %
                      (str(childName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsNamed(childName), recursive, showingOnly=showingOnly)


    def tab(self, tabName, recursive=True, showingOnly=None):
        """
        Search below this node for a tab with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("tab(self, tabName=%s, recursive=%s, showingOnly=%s)" %
                      (str(tabName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsATabNamed(tabName=tabName), recursive, showingOnly=showingOnly)


    def getUserVisibleStrings(self):
        """
        Get all user-visible strings in this node and its descendents.

        (Could be implemented as an attribute)
        """

        debug_log("getUserVisibleStrings(self)")

        result = []
        if self.name:
            result.append(self.name)

        if self.description:
            result.append(self.description)

        try:
            children = self.children
        except Exception:
            return result

        for child in children:
            result.extend(child.getUserVisibleStrings())

        return result


    def blink(self):
        """
        Blink will hightlight the current node on the screen. Using Blinker from utility.
        """

        debug_log("blink(self)")

        if not self.extents:
            return False
        else:
            (x, y, w, h) = self.extents
            Blinker(x, y, w, h)
            return True


class LinkAnchor(object):
    """
    Class storing info about an anchor within an Accessibility.Hyperlink, which
    is in turn stored within an Accessibility.Hypertext.
    """

    def __init__(self, node, hypertext, linkIndex, anchorIndex):
        self.node = node
        self.hypertext = hypertext
        self.linkIndex = linkIndex
        self.anchorIndex = anchorIndex


    @property
    def link(self):
        return self.hypertext.getLink(self.linkIndex)


    @property
    def URI(self):
        return self.link.getURI(self.anchorIndex)


class Root(Node):
    """
    Root class used to get data from Accessible.
    """

    def applications(self):
        """
        Get all applications.
        """

        debug_log("applications(self)")

        return root.findChildren(predicate.GenericPredicate(roleName="application"), recursive=False, showingOnly=False)


    def application(self, appName, retry=True):
        """
        Gets an application by name, returning an Application instance
        or raising an exception.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("applications(self, appName=%s, retry=%s)" % (appName, str(retry)))

        return root.findChild(predicate.IsAnApplicationNamed(appName), recursive=False, retry=retry, showingOnly=False)


class Application(Node):
    """
    Application class used to get data from Accessible.
    """

    def dialog(self, dialogName, recursive=False, showingOnly=None):
        """
        Search below this node for a dialog with the given name,
        returning a Window instance.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("dialog(self, dialogName=%s, recursive=%s, showingOnly=%s)" %
                      (str(dialogName), str(recursive), str(showingOnly)))

        return self.findChild(predicate.IsADialogNamed(dialogName=dialogName), recursive, showingOnly=showingOnly)


    def window(self, windowName, recursive=False, showingOnly=None):
        """
        Search below this node for a window with the given name,
        returning a Window instance.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """

        debug_log("window(self, windowName=%s, recursive=%s, showingOnly=%s)" %
                      (str(windowName), str(recursive), str(showingOnly)))

        result = self.findChild(predicate.IsAWindowNamed(windowName=windowName), recursive, showingOnly=showingOnly)
        return result


    def getWnckApplication(self, showingOnly=None):  # pragma: no cover
        """
        Get the wnck.Application instance for this application, or None

        Currently implemented via a hack: requires the app to have a
        window, and looks up the application of that window

        wnck.Application can give you the pid, the icon, etc
        """

        debug_log("getWnckApplication(self, showingOnly=%s) - untested" % str(showingOnly))

        window = self.child(roleName="frame", showingOnly=showingOnly)
        if window:
            wnckWindow = window.getWnckWindow()
            return wnckWindow.get_application()


class Window(Node):
    """
    Window class used to get data from Accessible.
    """

    def getWnckWindow(self):  # pragma: no cover
        """
        Get the wnck.Window instance for this window, or None
        """

        debug_log("getWnckWindow(self) - untested")

        screen = Wnck.screen_get_default()
        screen.force_update()
        for wnckWindow in screen.get_windows():
            if wnckWindow.get_name() == self.name:
                return wnckWindow


    def activate(self):  # pragma: no cover
        """
        Activates the wnck.Window associated with this Window.
        """

        debug_log("activate(self) - untested")

        wnckWindow = self.getWnckWindow()
        wnckWindow.activate(0)


class Wizard(Window):
    """
    Note that the buttons of a GnomeDruid were not accessible until recent versions of libgnomeui.
    This is http://bugzilla.gnome.org/show_bug.cgi?id=157936 and is fixed in
    gnome-2.10 and gnome-2.12 (in CVS libgnomeui) there's a patch attached to that bug.
    This bug is known to affect FC3; fixed in FC5
    """

    def __init__(self, node, debugName=None):
        Node.__init__(self, node)
        if debugName:
            self.debugName = debugName

        logger.log("%s is on '%s' page" % (self, self.getPageTitle()))


    def currentPage(self):
        """
        Get the current page of this wizard.
        This is currently a hack, supporting only GnomeDruid
        """

        debug_log("currentPage(self) - untested")

        pageHolder = self.child(roleName="panel")
        for child in pageHolder.children:
            if child.showing:
                return child

        raise "Unable to determine current page of %s" % self


    def getPageTitle(self):
        """
        Get the string title of the current page of this wizard
        This is currently a total hack, supporting only GnomeDruid
        """

        debug_log("getPageTitle(self) - untested")

        currentPage = self.currentPage()
        return currentPage.child(roleName="panel").child(roleName="panel").child(roleName="label", recursive=False).text


    def clickForward(self):
        """
        Click on the 'Forward' button to advance to next page of wizard.
        It will log the title of the new page that is reached.
        This will only work if your libgnomeui has accessible buttons see above.
        """

        debug_log("clickForward(self) - untested")

        fwd = self.child("Forward")
        fwd.click()

        logger.log("%s is now on '%s' page"% (self, self.getPageTitle()))


    def clickApply(self):
        """
        Click on the 'Apply' button to advance to next page of wizard.
        What if it's Finish rather than Apply?
        This will only work if your libgnomeui has accessible buttons see above.
        """

        debug_log("clickApply(self) - untested")

        fwd = self.child("Apply")
        fwd.click()


class Extension:
    @property
    def index_in_parent(self):
        for index, child in enumerate(self.get_parent().children):
            if child.get_parent().get_child_at_index(index) == self:
                return index

    @property
    def last_child(self):
        return not self.get_parent() or \
            self.index_in_parent is None or \
            self.index_in_parent == self.get_parent().get_child_count() - 1


    def _tree_branching(self, parent_prefix, was_parent_last, node, current, cutoff):
        if current == cutoff:
            return

        prefix_spacer = "     "
        prefix_extend = "  │  "
        suffix_branch = "  ├──"
        suffix_last =   "  └──"

        new_prefix = parent_prefix
        new_suffix = ""

        if node != self:
            if node.last_child:
                new_suffix = suffix_last
            else:
                new_suffix = suffix_branch

        if node != self and node.parent != self:
            if was_parent_last:
                current_prefix = prefix_spacer
            else:
                current_prefix = prefix_extend
            new_prefix += current_prefix


        node_data = "".join((
            f"[{node.name}-",
            f"{node.roleName}-",
            f"{node.description}]"
        ))

        print(f"{new_prefix+new_suffix} {node_data}")

        for child in node.children:
            self._tree_branching(new_prefix,
                                 was_parent_last=node.last_child,
                                 node=child,
                                 current=current+1,
                                 cutoff=cutoff)


    def tree(self, cutoff=None):
        return self._tree_branching(parent_prefix="",
                                    was_parent_last=False,
                                    node=self,
                                    current=0,
                                    cutoff=cutoff)


Accessibility.Accessible.__bases__ = (
    Application, Root, Node, Extension, ) + Accessibility.Accessible.__bases__


try:
    root = pyatspi.Registry.getDesktop(0)
    root.debugName = "root"
except Exception:  # pragma: no cover
    logger.log("Error: AT-SPI's desktop is not visible. Is accessibility enabled?")


children = root.children
if not children:  # pragma: no cover
    logger.log("Warning: AT-SPI's desktop is visible but it has no children. \
Are you running any AT-SPI-aware applications?")
del children


"""
Sniff also imports from tree and we don't want to run this code from sniff itself.
May have already been locked by dogtail.procedural.
Mark newly opened sniff not to use auto-refresh while script using this module is running.
"""
if not os.path.exists("/tmp/sniff_running.lock") or not os.path.exists("/tmp/sniff_refresh.lock"):
    sniff_lock = Lock(lockname="sniff_refresh.lock", randomize=False, unlockOnExit=True)
    try:
        sniff_lock.lock()
    except OSError:  # pragma: no cover
        pass

elif "sniff" not in sys.argv[0]:
    print("Dogtail: Warning: Running sniff has been detected.")
    print("Please make sure sniff has the 'Auto Refresh' disabled.")
    print("NOTE: Running scripts with sniff present is not recommended.")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from dogtail.config import config
from dogtail.utils import doDelay
from dogtail.logging import debug_log
from dogtail.logging import debugLogger as logger
from pyatspi import Registry as registry
from pyatspi import (KEY_SYM, KEY_PRESS, KEY_PRESSRELEASE, KEY_RELEASE)
from time import sleep
import os

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk

SESSION_TYPE = "x11"
if "XDG_SESSION_TYPE" in os.environ and "wayland" in os.environ["XDG_SESSION_TYPE"]:
    os.system("gsettings set org.gnome.shell introspect true")
    os.system("gsettings set org.gnome.desktop.peripherals.keyboard delay 2000")
    SESSION_TYPE = "wayland"

ponytail = None
if SESSION_TYPE == "wayland":
    from ponytail.ponytail import Ponytail
    ponytail = Ponytail()
    ponytail.disconnect()
    sleep(1) # needed now to overcome fast reconnect issue in ponytail

"""
Handles raw input using AT-SPI event generation.

Note: Think of keyvals as keysyms, and keynames as keystrings.
"""

__author__ = """
David Malcolm <dmalcolm@redhat.com>,
Zack Cerza <zcerza@redhat.com>
"""


def update_coords(coords):
    # Reserved for wrong coordinates reporting.
    return coords


# Detect Xwayland windows in order to use globals and recordMonitor
def ponytail_check_is_xwayland(window_id=None, window_list=None):
    if window_list is None:
        window_list = ponytail.window_list

    if isinstance(window_id, int):
        debug_log("Return wanted window_id.")
        return int([x["client-type"] for x in window_list if x["id"] == window_id][0])

    if window_id is None:
        debug_log("Return focused window if no window_id was provided.")
        for window in window_list:
            if bool(window["has-focus"]) is True:
                return int(window["client-type"])

    debug_log("Return window_id 0 which is gnome-shell if no id is found.")
    return 0


def ponytail_check_connection(window_id=None, input_source="mouse"):
    window_list = ponytail.window_list

    if (not isinstance(window_id, str)) and (window_list != []) and\
        window_id not in [x["id"] for x in window_list]:
        window_id = None

    debug_log("Checking if possibly connected window still exists.")

    if ponytail.connected and isinstance(ponytail.connected, int):
        if ponytail.connected not in [x["id"] for x in window_list]:
            debug_log("Disconnecting no longer open window.")
            ponytail.disconnect()
            sleep(1)
            debug_log("Done.")

    if input_source == "keyboard" and ponytail.connected is None:
        debug_log("Keyboard event, connecting monitor.")
        ponytail.connectMonitor()

    elif input_source == "keyboard" and ponytail.connected is not None:
        if window_id == "" and isinstance(ponytail.connected, int):
            debug_log("Keyboard event, monitor request, forcing monitor.")
            ponytail.disconnect()
            sleep(1)
            ponytail.connectMonitor()
        else:
            debug_log("Any window/monitor already connected for keyboard event.")

    else:
        debug_log("Mouse input event.")
        if ponytail_check_is_xwayland(window_id, window_list):
            window_id = ""

        if ponytail.connected is None and window_id is None:
            for window in window_list:
                if bool(window["has-focus"]) is True:
                    ponytail.connectWindow(window["id"])
                    debug_log("Connected active window.")

        elif ponytail.connected is not None and window_id is None:
            debug_log("Re-connected active window.")
            for window in window_list:
                if bool(window["has-focus"]) is True:
                    if ponytail.connected != window["id"]:
                        ponytail.disconnect()
                        sleep(1)
                        ponytail.connectWindow(window["id"])
                        debug_log("Re-connected active window.")

        elif ponytail.connected is None:
            if isinstance(window_id, int):
                ponytail.connectWindow(window_id)
                debug_log("Connected window by window_id.")

            elif isinstance(window_id, str):
                ponytail.connectMonitor(window_id)
                debug_log("Connected monitor (Xwayand?) by window_id.")

        elif ponytail.connected != window_id and isinstance(window_id, int):
            debug_log("Disconnecting window.")
            ponytail.disconnect()
            sleep(1.5)
            debug_log("Connected window: %s." % ponytail.connected)
            debug_log("Reconnecting window: %s." % window_id)
            ponytail.connectWindow(window_id)
            debug_log("Connected window: %s." % ponytail.connected)

        elif ponytail.connected != window_id and isinstance(window_id, str):
            debug_log("Disconnecting monitor.")
            ponytail.disconnect()
            sleep(1)
            debug_log("Reconnecting monitor.")
            ponytail.connectMonitor()
            debug_log("Connected monitor.")

        elif ponytail.connected == window_id:
            debug_log("Window or monitor is already connected.")


def doTypingDelay():
    doDelay(config.typingDelay)


def checkCoordinates(x, y):
    if x < 0 or y < 0:
        raise ValueError("Attempting to generate a mouse event at negative coordinates: (%s,%s)" % (x, y))


def click(x, y, button=1, check=True, window_id=None):
    """
    Synthesize a mouse button click at (x,y)
    """

    (x, y) = update_coords((x, y))

    if check:
        checkCoordinates(x, y)

    logger.log("Mouse button %s click at (%s,%s)" % (button, x, y))

    if SESSION_TYPE == "x11":
        registry.generateMouseEvent(x, y, name="b%sc" % button)

    else:
        ponytail_check_connection(window_id)
        ponytail.generateButtonEvent(button, x, y)

    doDelay(config.actionDelay)


def point(x, y, check=True, window_id=None):
    """
    Synthesize a button point at (x,y)
    """

    (x, y) = update_coords((x, y))

    if check:
        checkCoordinates(x, y)

    logger.log("Pointing mouse cursor at (%s,%s)" % (x, y))

    if SESSION_TYPE == "x11":
        registry.generateMouseEvent(x, y, "abs")

    else:
        ponytail_check_connection(window_id)
        ponytail.generateMotionEvent(x, y)

    doDelay(config.actionDelay)


def doubleClick(x, y, button=1, check=True, window_id=None):
    """
    Synthesize a mouse button double-click at (x,y)
    """

    (x, y) = update_coords((x, y))

    if check:
        checkCoordinates(x, y)

    logger.log("Mouse button %s doubleclick at (%s,%s)" % (button, x, y))

    if SESSION_TYPE == "x11":
        registry.generateMouseEvent(x, y, name="b%sd" % button)

    else:
        ponytail_check_connection(window_id)
        ponytail.generateButtonEvent(button, x, y)
        doDelay(config.doubleClickDelay)

        # Prevent executing release event on closed application
        ponytail.generateButtonPress(button)
        if ponytail.connected:
            ponytail.generateButtonRelease(button)
        else: # not ponytail.connected:
            ponytail.connectMonitor()
            ponytail.generateButtonRelease(button)

    doDelay(config.actionDelay)


def press(x, y, button=1, check=True, window_id=None, delay=config.defaultDelay):
    """
    Synthesize a mouse button press at (x,y)
    """

    (x, y) = update_coords((x, y))

    if check:
        checkCoordinates(x, y)

    logger.log("Mouse button %s press at (%s,%s)" % (button, x, y))

    if SESSION_TYPE == "x11":
        registry.generateMouseEvent(x, y, name="b%sp" % button)

    else:
        ponytail_check_connection(window_id)
        ponytail.generateMotionEvent(x, y)
        ponytail.generateButtonPress(button)

    doDelay(delay)


def release(x, y, button=1, check=True, window_id=None):
    """
    Synthesize a mouse button release at (x,y)
    """

    (x, y) = update_coords((x, y))

    if check:
        checkCoordinates(x, y)

    logger.log("Mouse button %s release at (%s,%s)" % (button, x, y))

    if SESSION_TYPE == "x11":
        registry.generateMouseEvent(x, y, name="b%sr" % button)

    else:
        ponytail_check_connection(window_id)
        ponytail.generateMotionEvent(x, y)
        doDelay()
        ponytail.generateButtonRelease(button)

    doDelay()


def absoluteMotion(x, y, mouseDelay=None, check=True, window_id=None):
    """
    Synthesize mouse absolute motion to (x,y)
    """

    (x, y) = update_coords((x, y))

    if check:
        checkCoordinates(x, y)

    logger.log("Mouse absolute motion to (%s,%s)" % (x, y))

    if SESSION_TYPE == "x11":
        registry.generateMouseEvent(x, y, name="abs")

    else:
        ponytail_check_connection(window_id)
        ponytail.generateMotionEvent(x, y)

    if mouseDelay:
        doDelay(mouseDelay)
    else:
        doDelay()


def absoluteMotionWithTrajectory(source_x, source_y, dest_x, dest_y, mouseDelay=None, check=True, window_id=None):
    """
    Synthetize mouse absolute motion with trajectory. The 'trajectory' means that the whole motion
    is divided into several atomic movements which are synthetized separately.
    """

    if check:
        checkCoordinates(source_x, source_y)
        checkCoordinates(dest_x, dest_y)

    logger.log("Mouse absolute motion with trajectory to (%s,%s)" % (dest_x, dest_y))

    if SESSION_TYPE == "wayland":
        ponytail_check_connection(window_id)

    dx = float(dest_x - source_x)
    dy = float(dest_y - source_y)
    max_len = max(abs(dx), abs(dy))

    if max_len == 0:
        return

    dx /= max_len
    dy /= max_len
    act_x = float(source_x)
    act_y = float(source_y)

    for _ in range(0, int(max_len)):
        act_x += dx
        act_y += dy
        if mouseDelay:
            doDelay(mouseDelay)

        if SESSION_TYPE == "x11":
            registry.generateMouseEvent(int(act_x), int(act_y), name="abs")

        else:
            ponytail.generateMotionEvent(int(act_x), int(act_y))

    if mouseDelay:
        doDelay(mouseDelay)
    else:
        doDelay()


def relativeMotion(x, y, mouseDelay=None):
    """
    Synthetize a relative motion from actual position.
    Note: Does not check if the end coordinates are positive.
    """

    logger.log("Mouse relative motion of (%s,%s)" % (x, y))

    if SESSION_TYPE == "wayland":
        logger.log("Relative motion unavailable under wayland not available.")
        return

    registry.generateMouseEvent(x, y, name="rel")

    if mouseDelay:
        doDelay(mouseDelay)
    else:
        doDelay()


def drag(fromXY, toXY, button=1, check=True):
    """
    Synthesize a mouse press, drag, and release on the screen.
    """

    logger.log("Mouse button %s drag with trajectory from %s to %s" % (button, fromXY, toXY))

    (x, y) = fromXY
    press(x, y, button, check)

    (x, y) = toXY
    absoluteMotion(x, y, check=check)
    doDelay()

    release(x, y, button, check)
    doDelay()


def dragNodeToNode(source_node, dest_node, button=1, check=True):
    """
    Drag source_node onto dest_node. These are tree.Node objects. Takes positions
    of these Nodes directly, so you don't have to calculate end enter them directly.
    """

    logger.log("Drag node to node source_node=%s, dest_node=%s, button=%s, check=%s" %
                  (str(source_node), str(dest_node), str(button), str(check)))

    x = source_node.position[0] + source_node.size[0] / 2
    y = source_node.position[1] + source_node.size[1] / 2
    press(x, y, button, check, source_node.window_id)

    x = dest_node.position[0] + dest_node.size[0] / 2
    y = dest_node.position[1] + dest_node.size[1] / 2

    release(x, y, button, check, window_id=dest_node.window_id)
    doDelay()


def dragWithTrajectoryGlobal(fromXY, toXY, button=1):
    """
    Synthetize a mouse press, drag (including move events), and release on the screen
    For use on Wayland - as this function forces using global coords, although we get
    local ones from a11y. So this function is targeted to be used for inter-window drags
    on Wayland, where you will need to questimate the coords. Which is doable i.e by having
    source window placed on the left *half* of the screen (local coords will be very similar
    to globals perhaps with the top panel offset).

    Having 'trajectory' appears to be necessary on Wayland for any drags. Use 'dragWithTrajectory'
    or just 'drag' on X sessions like in pre-wayland version of dogtail.
    """

    logger.log("Drag with trajectory global fromXY=%s, toXY=%s, button=%s" %
                  (str(fromXY), str(toXY), str(button)))

    if SESSION_TYPE == "wayland":
        (x, y) = fromXY
        ponytail_check_connection(window_id="")

        ponytail.generateMotionEvent(x, y)
        doDelay(config.defaultDelay)
        ponytail.generateButtonPress(1)
        doDelay(config.defaultDelay)

        (x, y) = toXY
        absoluteMotionWithTrajectory(fromXY[0], fromXY[1], x, y, mouseDelay=0.01, window_id="")

        ponytail.generateMotionEvent(x, y)
        doDelay(config.defaultDelay)
        ponytail.generateButtonRelease(1)
        doDelay()

    else:
        dragWithTrajectory(fromXY, toXY, button)


def dragWithTrajectory(fromXY, toXY, button=1, check=True, press_delay=config.defaultDelay, mouse_delay=0.01):
    """
    Synthetize a mouse press, drag (including move events), and release on the screen
    Please note, that on Wayland, this function works for drags only within a single window.
    On X this function works with global coords and equals dragWithTrajectoryGlobal
    """

    logger.log("Drag with trajectory fromXY=%s, toXY=%s, button=%s, check=%s, press_delay=%s, mouse_delay=%s" %
                  (str(fromXY), str(toXY), str(button), str(check), str(press_delay), str(mouse_delay)))

    (x, y) = fromXY
    press(x, y, button, check, delay=press_delay)

    (x, y) = toXY
    absoluteMotionWithTrajectory(fromXY[0], fromXY[1], x, y, mouseDelay=mouse_delay, check=check)
    doDelay()

    release(x, y, button, check)
    doDelay()


def typeText(string):
    """
    Types the specified string, one character at a time.
    Please note, you may have to set a higher typing delay,
    if your machine misses/switches the characters typed.
    Needed sometimes on slow setups/VMs typing non-ASCII utf8 chars.
    """

    debug_log("typeText(string=%s)" % string)

    for char in string:
        pressKey(char)


keyNameAliases = {
    "enter": "Return",
    "esc": "Escape",
    "backspace": "BackSpace",
    "alt": "Alt_L",
    "control": "Control_L",
    "ctrl": "Control_L",
    "shift": "Shift_L",
    "del": "Delete",
    "ins": "Insert",
    "pageup": "Page_Up",
    "pagedown": "Page_Down",
    "win": "Super_L",
    "meta": "Super_L",
    "super": "Super_L",
    "tab": "Tab",
    "print": "Print",
    " ": "space",
    "\t": "Tab",
    "\n": "Return",
    "\b": "BackSpac"
}


def uniCharToKeySym(uniChar):
    """
    Use GDK to get the key symbol for a given unicode character.
    """

    debug_log("uniCharToKeySym(uniChar=%s)" % uniChar)

    i = ord(uniChar)
    keySym = Gdk.unicode_to_keyval(i)
    return keySym


def keyNameToKeySym(keyName):
    """
    Use GDK to get the key symbol for a key name.
    """

    debug_log("keyNameToKeySym(keyName=%s)" % keyName)

    keyName = keyNameAliases.get(keyName.lower(), keyName)
    keySym = Gdk.keyval_from_name(keyName)
    # various error 'codes' returned for non-recognized chars in versions of GTK3.X
    if keySym == 0xffffff or keySym == 0x0 or keySym is None:
        try:
            keySym = uniCharToKeySym(keyName)
        except Exception:  # not even valid utf-8 char
            try:  # Last attempt run at a keyName ('Meta_L', 'Dash' ...)
                keySym = getattr(Gdk, "KEY_" + keyName)
            except AttributeError:
                raise KeyError(keyName)
    return keySym


def keyNameToKeyCode(keyName):
    """
    Use GDK to get the keycode for a given keystring.

    Note that the keycode returned by this function is often incorrect when
    the requested keystring is obtained by holding down the Shift key.

    Generally you should use uniCharToKeySym() and should only need this
    function for nonprintable keys anyway.
    """

    debug_log("keyNameToKeyCode(keyName=%s)" % keyName)

    keymap = Gdk.Keymap.get_for_display(Gdk.Display.get_default())
    entries = keymap.get_entries_for_keyval(Gdk.keyval_from_name(keyName))

    try:
        return entries[1][0].keycode
    except TypeError:
        pass


def pressKey(keyName, window_id=None):
    """
    Presses (and releases) the key specified by keyName.
    keyName is the English name of the key as seen on the keyboard. Ex: 'enter'
    Names are looked up in Gdk.KEY_ If they are not found there, they are
    looked up by uniCharToKeySym().
    """

    debug_log("pressKey(keyName=%s, window_id=%s)" % (keyName, str(window_id)))

    if keyName.lower() in ("esc", "escape", "enter", "return"):
        window_id = "" # when this would quit a window, release event would be doomed

    keySym = keyNameToKeySym(keyName)
    if SESSION_TYPE == "x11":
        registry.generateKeyboardEvent(keySym, None, KEY_SYM)
        doTypingDelay()

    else:
        ponytail_check_connection(window_id, input_source="keyboard")
        ponytail.generateKeysymEvent(keySym, delay=config.typingDelay/2)



def keyCombo(comboString):
    """
    Generates the appropriate keyboard events to simulate a user pressing the
    specified key combination.

    comboString is the representation of the key combo to be generated.
    e.g. '<Control><Alt>p' or '<Control><Shift>PageUp' or '<Control>q'
    """

    debug_log("keyCombo(comboString=%s)" % comboString)

    strings = []
    for s in comboString.split('<'):
        if s:
            for S in s.split('>'):
                if S:
                    S = keyNameAliases.get(S.lower(), S)
                    strings.append(S)

    for s in strings:
        if not hasattr(Gdk, s):
            if not hasattr(Gdk, "KEY_" + s):
                raise ValueError("Cannot find key '%s'" % s)

    modifiers = strings[:-1]
    finalKey = strings[-1]

    if SESSION_TYPE == "x11":
        for modifier in modifiers:
            code = keyNameToKeyCode(modifier)
            registry.generateKeyboardEvent(code, None, KEY_PRESS)

        code = keyNameToKeyCode(finalKey)
        registry.generateKeyboardEvent(code, None, KEY_PRESSRELEASE)

        for modifier in modifiers:
            code = keyNameToKeyCode(modifier)
            registry.generateKeyboardEvent(code, None, KEY_RELEASE)

    else:
        # always use monitor, window will often get closed
        # before final release i.e. with alt-f4 ctrl-q etc!
        ponytail_check_connection(input_source="keyboard", window_id="")

        for modifier in modifiers:
            code = keyNameToKeyCode(modifier)
            ponytail.generateKeycodePress(code)

        code = keyNameToKeyCode(finalKey)
        ponytail.generateKeycodeEvent(code)

        for modifier in modifiers:
            code = keyNameToKeyCode(modifier)
            ponytail.generateKeycodeRelease(code)

    doDelay()


def holdKey(keyName):
    """
    Press and hold the key specified by keyName.
    """

    debug_log("holdKey(keyName=%s)" % keyName)

    keyName = keyNameAliases.get(keyName.lower(), keyName)
    code = keyNameToKeyCode(keyName)

    if SESSION_TYPE == "x11":
        registry.generateKeyboardEvent(code, None, KEY_PRESS)

    else:
        ponytail_check_connection(input_source="keyboard")
        ponytail.generateKeycodePress(code)

    doDelay()


def releaseKey(keyName):
    """
    Releases the holded key specified by keyName.
    """

    debug_log("releaseKey(keyName=%s)" % keyName)

    keyName = keyNameAliases.get(keyName.lower(), keyName)
    code = keyNameToKeyCode(keyName)

    if SESSION_TYPE == "x11":
        registry.generateKeyboardEvent(code, None, KEY_RELEASE)

    else:
        ponytail_check_connection(input_source="keyboard")
        ponytail.generateKeycodeRelease(code)

    doDelay()



# -*- coding: utf-8 -*-
import os
from dogtail import tree
from dogtail.utils import run
from dogtail.predicate import GenericPredicate


def gedit_text_tree_api_demo():
    os.environ["LANG"] = "en_US.UTF-8"

    # Remove the output file, if it's still there from a previous run
    if os.path.isfile(os.path.join("/tmp", "UTF8demo.txt")):
        os.remove(os.path.join("/tmp", "UTF8demo.txt"))

    # Start gedit.
    # Get a handle to gedit's application object.
    os.system("killall gedit") # preventing the undesired result if the sript is started multiple times in a row
    run("gedit")
    gedit = tree.root.application("gedit")

    # Get a handle to gedit's text object.
    text_buffer = gedit.findChildren(GenericPredicate(roleName="text"))[1]
    # is equal to
    # text_buffer = gedit.findChild(lambda x: x.roleName=="text" and x.showing) # we can specify search by another attribute - showing
    # Load the UTF-8 demo file.
    # Set the attribute to the given text object
    with open(os.path.abspath(".") + "/data/UTF-8-demo.txt", "r") as open_file:
        text_buffer.text = open_file.read()


    # Get a handle to gedit's Save button and click.
    save_button = gedit.findChild(lambda x: x.name == "Save" and x.roleName in ("button", "push button"))
    save_button.click()


    # We want to save to the file name 'UTF8demo.txt'.
    dialog = gedit.findChild(lambda x: "Save As" in x.name and x.roleName == "file chooser")
    dialog.child(roleName="text").text = "/tmp/UTF8demo.txt"


    # Get a handle to gedit's Save dialog and click.
    save_as_button = gedit.findChildren(lambda x: "Save" in x.name)[-1]
    save_as_button.click()


    # Let's quit now.
    gedit.child("Menu", "toggle button").click()
    # Sometimes the node is not unique so you need to specify another attribute to look for
    gedit.findChild(lambda x: x.name == "Close" and x.showing).click()


gedit_text_tree_api_demo()
